# Copyright (c) 2003, James Seng. (http://james.seng.cc/)
# This code is released under the Artistic License.

package MT::Bayesian;
use strict;

use MT::Object;
use MT::BayesianBlog;
use MT::BayesianToken;
use MT::Comment;
use MT::TBPing;

@MT::Bayesian::ISA = qw( MT::Object );

###############################
#                             #
# Define the follow variables #
#                             #
###############################

# tokens which is too common to be used to determine if it is spam or not
my @ignore_tokens = qw/ftp http smtp www com net org biz info this spam like/;

# consider blacklist or whitelisting on a per host or per subnet basis
my $ban_subnet = 1;

###########################
#                         #
# Do not modify from here #
#                         #
###########################

# min token length for it to be taken into consideration
my $min_token_length = 3;

# max token length for it to be taken into consideration
my $max_token_length = 15;

# min token training before it is counted
my $min_training = 5;

my $significant = 20;
my $threshold_spam = 0.9;
my $threshold_notspam = 0.2;
my $fudgefactor = 2;

__PACKAGE__->install_properties({
    columns => [
        'id', 'blog_id', 'ping_id', 'comment_id', 
        'spam', 'prob', 'is_trained', 'is_tested',
    ],
    indexes => {
        blog_id => 1,
        comment_id => 1,
        ping_id => 1,
    },
    datasource => 'bayesian',
    primary_key => 'id',
});

sub _max {
  my @t=@_;
  my $a=$t[0];
  foreach my $b (@t[1..$#t]) {
    if ($b>$a) {
      $a=$b;
    }
  }
  return $a;
}

sub _min {
  my @t=@_;
  my $a=$t[0];
  foreach my $b (@t[1..$#t]) {
    if ($b<$a) {
      $a=$b;
    }
  }
  return $a;
}

sub _lower {
    my $a=shift;
    $a =~ tr/A-Z/a-z/;
    return $a;
}

sub new_bayesian {
    my $self = shift;
    my $hash = shift;

    foreach my $k (keys %$hash) {
        $self->$k($hash->{$k});
    }
    $self->prob(0.5);
    $self->spam(0);
    $self->test_spam(1);
}

sub _tokenise {
    my ($self,$message) = @_;

    my @tokens; my $token;

    return @tokens if (!$message || $message eq '');

    # find generic token
    foreach $token (split /[^-\$A-Za-z0-9\']+/,$message) {
        # ignore empty token
        next if ($token eq '');

        # ignore tokens which is in a special list
        my $ignore = 0;
        foreach (@ignore_tokens) {
            $ignore = 1, last if $token eq $_;
        }
        next if ($ignore);

        # ignore all numbers token
        next if ($token =~ /^[0-9]+$/); 

        # ignore if token is too short
        next if (length($token) < $min_token_length); 

        # ignore if token is too long
        next if (length($token) > $max_token_length); 

        push @tokens,_lower($token);
    }

    # find the hostnames in URL
    foreach $token ($message =~ /\:\/\/([A-Za-z0-9\_\-\.]+)/g)  {
        $token =~ s/^www//i;        # ignore the leading www
        push @tokens, _lower($token);
    }

    # find email addresses
    foreach my $token 
        ($message =~ /([A-Za-z0-9\_\-\.\+]+\@[A-Za-z0-9_\-\.]+)/g) {
        push @tokens, _lower($token);
    }

    # find ip address
    foreach my $token ($message =~ /(\d+\.\d+\.\d+.\d+)/g) {
        $token =~ s/\.\d+$// if ($ban_subnet);  # consider per subnet
        push @tokens, _lower($token);
    }

    return @tokens;
}

sub train_spam {
    my $self = shift;
    my $message;
    my $bt;

    if ($self->is_trained) {
        return if ($self->is_trained == 1);
        $self->untrain;
    }

    if ($self->comment_id) {
        use MT::Comment;

        my $cm = MT::Comment->load($self->comment_id);
        return if (!$cm);

        $message = join(' ', $cm->text, $cm->ip, $cm->url, $cm->email);
    } elsif ($self->ping_id) {
        use MT::TBPing;

        my $ping = MT::TBPing->load($self->ping_id);
        return if (!$ping);

        $message = join(' ', $ping->title, $ping->blog_name, 
                             $ping->excerpt, $ping->ip,
                             $ping->source_url);
    }

    my @tokens = $self->_tokenise($message);

    my $by = MT::BayesianBlog->load({blog_id => $self->blog_id});
    if (!$by) {
        $by = new MT::BayesianBlog;
        $by->blog_id($self->blog_id);
        $by->total_spam(0);
        $by->total_nonspam(0);
    }
    $by->total_spam($by->total_spam+1);
    $by->save;

    $self->spam(1);
    $self->prob(1);
    $self->is_trained(1);
    $self->save;

    foreach my $t (@tokens) {
        undef $bt;
        $bt = MT::BayesianToken->load({ blog_id => $self->blog_id,
                                           spam => 1,
                                           token => $t }); 
        if (!$bt) {
            $bt = MT::BayesianToken->new;
            $bt->blog_id($self->blog_id);
            $bt->spam(1);
            $bt->token($t);
            $bt->token_count(1);
        } else {
            $bt->token_count($bt->token_count+1);
        }
        $bt->save;
    }
}

sub train_nonspam {
    my $self = shift;
    my $message;
    my $bt;

    if ($self->is_trained) {
        return if ($self->is_trained == -1);
        $self->untrain;
    }

    if ($self->comment_id) {
        use MT::Comment;

        my $cm = MT::Comment->load($self->comment_id);
        return if (!$cm);

        $message = join(' ', $cm->text, $cm->ip, $cm->url, $cm->email);
    } elsif ($self->ping_id) {
        use MT::TBPing;

        my $ping = MT::TBPing->load($self->ping_id);
        return if (!$ping);

        $message = join(' ', $ping->title, $ping->blog_name, 
                             $ping->excerpt, $ping->ip,
                             $ping->source_url);
    }

    my @tokens = $self->_tokenise($message);

    my $by = MT::BayesianBlog->load({blog_id => $self->blog_id});
    if (!$by) {
        $by = new MT::BayesianBlog;
        $by->blog_id($self->blog_id);
        $by->total_spam(0);
        $by->total_nonspam(0);
    }
    $by->total_nonspam($by->total_nonspam+1);
    $by->save;

    $self->spam(-1);
    $self->prob(0);
    $self->is_trained(-1);
    $self->save;

    foreach my $t (@tokens) {
        undef $bt;
        $bt = MT::BayesianToken->load({ blog_id => $self->blog_id,
                                           spam => -1,
                                           token => $t }); 
        if (!$bt) {
            $bt = MT::BayesianToken->new;
            $bt->blog_id($self->blog_id);
            $bt->spam(-1);
            $bt->token($t);
            $bt->token_count(1);
        } else {
            $bt->token_count($bt->token_count+1);
        }
        $bt->save;
    }

}

sub untrain {
    my $self = shift;
    my $message;

    # cannot untrain if it is not trained before
    return if (!$self->is_trained);

    # if cannot find the blog entry, then it cannot have been trained
    my $by = MT::BayesianBlog->load({blog_id => $self->blog_id});
    return if (!$by);

    if ($self->comment_id) {
        use MT::Comment;

        my $cm = MT::Comment->load($self->comment_id);
        return if (!$cm);

        $message = join(' ', $cm->text, $cm->ip, $cm->url, $cm->email);
    } elsif ($self->ping_id) {
        use MT::TBPing;

        my $ping = MT::TBPing->load($self->ping_id);
        return if (!$ping);

        $message = join(' ', $ping->title, $ping->blog_name, 
                             $ping->excerpt, $ping->ip,
                             $ping->source_url);
    }

    # reduce the spam count
    $by->total_spam($by->total_spam-1);
    $by->save;

    # untrain the message
    my $bt;
    my @tokens = $self->_tokenise($message);
    foreach my $t (@tokens) {
        undef $bt;
        $bt = MT::BayesianToken->load({ blog_id => $self->blog_id,
                                           spam => $self->is_trained,
                                           token => $t }); 
        next if (!$bt);
        $bt->token_count($bt->token_count-1);
        $bt->save;
    }

    $self->spam(0);
    $self->prob(0);
    $self->is_trained(0);
    $self->save;
}

sub test_spam {
    my $self = shift;
    my $force = shift;
    my $message;
    my $count = 0;

    # if the entry is trained, then just use the trained result
    return ($self->spam, $self->prob) if ($self->is_trained);

    # if the entry is tested before and we are not forcing a recalculation
    # dont recalculate
    return ($self->spam, $self->prob) if (!$force && $self->is_tested);

    # default values;
    my $spam = 0;
    my $prob = 0.5;
    my $total_spam = 1;
    my $total_nonspam = 1;
    my %tokencache;

    if ($self->comment_id) {
        use MT::Comment;

        my $cm = MT::Comment->load($self->comment_id);
        return ($spam,$prob) if (!$cm);
        $message = join(' ', $cm->text, $cm->ip, $cm->url, $cm->email);
    } elsif ($self->ping_id) {
        use MT::TBPing;

        my $ping = MT::TBPing->load($self->ping_id);
        return ($spam,$prob) if (!$ping);

        $message = join(' ', $ping->title, $ping->blog_name, 
                             $ping->excerpt, $ping->ip,
                             $ping->source_url);
    } else {
        return (0,0.5);
    }

    my @tokens = $self->_tokenise($message);

    my $by = MT::BayesianBlog->load({blog_id => $self->blog_id});
    if ($by) {
        $total_spam = &_max(1,$by->total_spam);
        $total_nonspam = &_max(1,$by->total_nonspam);
    }

    foreach my $token (@tokens) {
        my $bt;
        my $token_spam = 0;
        my $token_nonspam = 0;

        $tokencache{$token} = 0.5;
        $count++;

        $bt = MT::BayesianToken->load({ blog_id => $self->blog_id,
                                           spam => -1,
                                           token => $token }); 
        $token_nonspam = 1,$bt->token_count if ($bt);

        $bt = MT::BayesianToken->load({ blog_id => $self->blog_id,
                                           spam => 1,
                                           token => $token }); 
        $token_spam =$bt->token_count if ($bt);

        # take special note of hostname, ip, url, and emails
        my $c = ($token =~ /\./) ? 0 : $min_training;

        $token_nonspam *= $fudgefactor;
        if ($token_spam + $token_nonspam > $c) {
            $tokencache{$token} = 
              &_max(0.01,&_min(0.99,
                &_min($token_spam/$total_spam,1)/
                  (&_min($token_nonspam/$total_nonspam,1)+
                   &_min($token_spam/$total_spam,1))
              ));
        } 

        #printf "Atoken %s,%f,%f\n",$token,$tokencache{$token},$prob;

        if (!$c) {
            $prob *= ($tokencache{$token} /
                      ($prob*$tokencache{$token} 
                      + ((1-$prob) * (1-$tokencache{$token}))));

        #printf "Btoken %s,%f,%f\n",$token,$tokencache{$token},$prob;
            $prob *= ($tokencache{$token} /
                      ($prob*$tokencache{$token} 
                      + ((1-$prob) * (1-$tokencache{$token}))));

        #printf "Btoken %s,%f,%f\n",$token,$tokencache{$token},$prob;
        }

    }

    my @toklist = sort 
        {abs($tokencache{$b}-0.5) <=> abs($tokencache{$a}-0.5)} @tokens;

    my $i;
    for ($i=0;$i<$significant;$i++) {
        last if ($i>=$count);

        $prob *= ($tokencache{$toklist[$i]} /
            ($prob*$tokencache{$toklist[$i]} 
            + ((1-$prob) * (1-$tokencache{$toklist[$i]}))));
    }

    $prob = 0 if ($prob < 0);
    $prob = 1 if ($prob > 1);

    if ($prob > $threshold_spam) {
        $spam = 1;
    } elsif ($prob < $threshold_notspam) {
        $spam = -1;
    }

    $self->spam($spam);
    $self->prob($prob);
    $self->is_tested(1);
    $self->save;

    return ($spam, $prob);
}

1;
